@@ -78,6 +78,8 @@ module Sinatra
       super
     rescue Rack::Utils::ParameterTypeError, Rack::Utils::InvalidParameterError => e
       raise BadRequest, "Invalid query parameters: #{Rack::Utils.escape_html(e.message)}"
+    rescue EOFError => e
+      raise BadRequest, "Invalid multipart/form-data: #{Rack::Utils.escape_html(e.message)}"
     end
 
     class AcceptEntry
@@ -188,7 +190,7 @@ module Sinatra
         headers["Content-Length"] = body.map(&:bytesize).reduce(0, :+).to_s
       end
 
-      [status.to_i, headers, result]
+      [status, headers, result]
     end
 
     private
@@ -198,11 +200,11 @@ module Sinatra
     end
 
     def drop_content_info?
-      status.to_i / 100 == 1 or drop_body?
+      informational? or drop_body?
     end
 
     def drop_body?
-      DROP_BODY_RESPONSES.include?(status.to_i)
+      DROP_BODY_RESPONSES.include?(status)
     end
   end
 
@@ -869,12 +871,12 @@ module Sinatra
 
     def compile_template(engine, data, options, views)
       eat_errors = options.delete :eat_errors
-      template_cache.fetch engine, data, options, views do
-        template = Tilt[engine]
-        raise "Template engine not found: #{engine}" if template.nil?
+      template = Tilt[engine]
+      raise "Template engine not found: #{engine}" if template.nil?
 
-        case data
-        when Symbol
+      case data
+      when Symbol
+        template_cache.fetch engine, data, options, views do
           body, path, line = settings.templates[data]
           if body
             body = body.call if body.respond_to?(:call)
@@ -892,17 +894,24 @@ module Sinatra
             throw :layout_missing if eat_errors and not found
             template.new(path, 1, options)
           end
-        when Proc, String
-          body = data.is_a?(String) ? Proc.new { data } : data
-          caller = settings.caller_locations.first
-          path = options[:path] || caller[0]
-          line = options[:line] || caller[1]
-          template.new(path, line.to_i, options, &body)
-        else
-          raise ArgumentError, "Sorry, don't know how to render #{data.inspect}."
         end
+      when Proc
+        compile_block_template(template, options, &data)
+      when String
+        template_cache.fetch engine, data, options, views do
+          compile_block_template(template, options) { data }
+        end
+      else
+        raise ArgumentError, "Sorry, don't know how to render #{data.inspect}."
       end
     end
+
+    def compile_block_template(template, options, &body)
+      caller = settings.caller_locations.first
+      path = options[:path] || caller[0]
+      line = options[:line] || caller[1]
+      template.new(path, line.to_i, options, &body)
+    end
   end
 
   # Base class for all Sinatra applications and middleware.
@@ -916,7 +925,7 @@ module Sinatra
     attr_accessor :app, :env, :request, :response, :params
     attr_reader   :template_cache
 
-    def initialize(app = nil)
+    def initialize(app = nil, **kwargs)
       super()
       @app = app
       @template_cache = Tilt::Cache.new
@@ -934,6 +943,7 @@ module Sinatra
       @params   = IndifferentHash.new
       @request  = Request.new(env)
       @response = Response.new
+      @pinned_response = nil
       template_cache.clear if settings.reload_templates
 
       invoke { dispatch! }
@@ -994,11 +1004,11 @@ module Sinatra
 
     # Run filters defined on the class and all superclasses.
     # Accepts an optional block to call after each filter is applied.
-    def filter!(type, base = settings)
-      filter! type, base.superclass if base.superclass.respond_to?(:filters)
+    def filter!(type, base = settings, &block)
+      filter!(type, base.superclass, &block) if base.superclass.respond_to?(:filters)
       base.filters[type].each do |args|
         result = process_route(*args)
-        yield result if block_given?
+        block.call(result) if block_given?
       end
     end
 
@@ -1006,7 +1016,7 @@ module Sinatra
     def route!(base = settings, pass_block = nil)
       if routes = base.routes[@request.request_method]
         routes.each do |pattern, conditions, block|
-          @response.delete_header('Content-Type') unless @pinned_response
+          response.delete_header('Content-Type') unless @pinned_response
 
           returned_pass_block = process_route(pattern, conditions) do |*args|
             env['sinatra.route'] = "#{@request.request_method} #{pattern}"
@@ -1089,6 +1099,7 @@ module Sinatra
       return unless valid_path?(path)
 
       path = File.expand_path(path)
+      return unless path.start_with?(File.expand_path(public_dir) + '/')
       return unless File.file?(path)
 
       env['sinatra.static_file'] = path
@@ -1124,7 +1135,7 @@ module Sinatra
       invoke do
         static! if settings.static? && (request.get? || request.head?)
         filter! :before do
-          @pinned_response = !@response['Content-Type'].nil?
+          @pinned_response = !response['Content-Type'].nil?
         end
         route!
       end
@@ -1145,7 +1156,7 @@ module Sinatra
       end
       @env['sinatra.error'] = boom
 
-      if boom.respond_to? :http_status
+      if boom.respond_to? :http_status and boom.http_status.between? 400, 599
         status(boom.http_status)
       elsif settings.use_code? and boom.respond_to? :code and boom.code.between? 400, 599
         status(boom.code)
@@ -1153,8 +1164,6 @@ module Sinatra
         status(500)
       end
 
-      status(500) unless status.between? 400, 599
-
       if server_error?
         dump_errors! boom if settings.dump_errors?
         raise boom if settings.show_exceptions? and settings.show_exceptions != :after_handler
@@ -1168,7 +1177,7 @@ module Sinatra
 
       if not_found? || bad_request?
         if boom.message && boom.message != boom.class.name
-          body boom.message
+          body Rack::Utils.escape_html(boom.message)
         else
           content_type 'text/html'
           body '<h1>' + (not_found? ? 'Not Found' : 'Bad Request') + '</h1>'
@@ -1442,7 +1451,7 @@ module Sinatra
       # in `extensions` available to the handlers and templates
       def helpers(*extensions, &block)
         class_eval(&block)   if block_given?
-        prepend(*extensions) if extensions.any?
+        include(*extensions) if extensions.any?
       end
 
       # Register an extension. Alternatively take a block from which an
@@ -1471,6 +1480,7 @@ module Sinatra
         @prototype = nil
         @middleware << [middleware, args, block]
       end
+      ruby2_keywords(:use) if respond_to?(:ruby2_keywords, true)
 
       # Stop the self-hosted server if running.
       def quit!
@@ -1490,7 +1500,7 @@ module Sinatra
       def run!(options = {}, &block)
         return if running?
         set options
-        handler         = detect_rack_handler
+        handler         = Rack::Handler.pick(server)
         handler_name    = handler.name.gsub(/.*::/, '')
         server_settings = settings.respond_to?(:server_settings) ? settings.server_settings : {}
         server_settings.merge!(:Port => port, :Host => bind)
@@ -1523,8 +1533,8 @@ module Sinatra
       # Create a new instance of the class fronted by its middleware
       # pipeline. The object is guaranteed to respond to #call but may not be
       # an instance of the class new was called on.
-      def new(*args, &bk)
-        instance = new!(*args, &bk)
+      def new(*args, **kwargs, &bk)
+        instance = new!(*args, **kwargs, &bk)
         Wrapper.new(build(instance).to_app, instance)
       end
 
@@ -1744,17 +1754,6 @@ module Sinatra
         builder.use session_store, options
       end
 
-      def detect_rack_handler
-        servers = Array(server)
-        servers.each do |server_name|
-          begin
-            return Rack::Handler.get(server_name.to_s)
-          rescue LoadError, NameError
-          end
-        end
-        fail "Server handler (#{servers.join(',')}) not found."
-      end
-
       def inherited(subclass)
         subclass.reset!
         subclass.set :app_file, caller_files.first unless subclass.app_file?
@@ -1958,6 +1957,8 @@ module Sinatra
           return super(*args, &block) if respond_to? method_name
           Delegator.target.send(method_name, *args, &block)
         end
+        # ensure keyword argument passing is compatible with ruby >= 2.7
+        ruby2_keywords(method_name) if respond_to?(:ruby2_keywords, true)
         private method_name
       end
     end
